package main

import (
	"fmt"
	"github.com/julienschmidt/httprouter"
	"gopkg.in/robfig/cron.v2"
	"hivedav/caldav"
	"hivedav/config"
	"log"
	"net/http"
	"time"
)

var hivedavVersion = "v0.2"

func main() {
	// If exists, load config file from pwd
	conf := &config.Config{}
	conf, err := conf.LoadConfig(".")
	if err != nil {
		log.Fatal("Error loading config: ", err)
	}

	var server *Server
	server, err = server.NewServer(conf)
	if err != nil {
		log.Fatal("Error creating server with config: ", err)
	}

	log.Printf("HiveDAV %s, %s 🍯\n", hivedavVersion, conf.HiveDavHost)
	log.Println("----------------------------------------------------")
	log.Printf("Cron interval:\t%d minutes\n", conf.RefreshInterval)
	log.Printf("CalDAV server:\t%s\n", conf.CaldavUri)
	log.Printf("CalDAV user:\t%s\n", conf.CaldavUser)

	c := cron.New()
	c.AddFunc(fmt.Sprintf("@every %dm", conf.RefreshInterval), func() { updateDatabase(server, c) })
	go updateDatabase(server, c)
	c.Start()

	router := httprouter.New()
	router.GET("/", server.Index)
	router.GET("/week/:week", server.Week)
	router.GET("/list", server.ListCubicles)
	router.GET("/list/:week", server.ListCubiclesInWeek)
	router.POST("/book/:dtstart", server.BookCubicle)
	router.GET("/book/:dtstart", server.CubicleForm)
	router.ServeFiles("/css/*filepath", http.Dir("./css"))
	router.GET("/healthz", server.Healthz)
	log.Printf("Listening:\t\t%s:%d\n", conf.ListenAddress, conf.ListenPort)
	log.Fatal(http.ListenAndServe(fmt.Sprintf("%s:%d", conf.ListenAddress, conf.ListenPort), router))
}

func updateDatabase(s *Server, c *cron.Cron) {
	// Find current user principal
	userPrincipal, err := caldav.UserPrincipal(s.config.CaldavUri, s.config.CaldavUser, s.config.CaldavPassword)
	if err != nil {
		log.Fatal("Error reading current user principal: ", err)
	}
	log.Printf("User principal:\t%s\n", userPrincipal)

	// Find home set of current user principal
	homeSetUri, err := caldav.CalendarHome(s.config.CaldavUri, s.config.CaldavUser, s.config.CaldavPassword, userPrincipal, s.config.Calendar)
	if err != nil {
		log.Fatal("Error reading home set URI: ", err)
	}
	log.Printf("Calendar homeset:\t%s", homeSetUri)

	// Find calendar URI for the calendar specified in the config
	calendarUri, err := caldav.CalendarFromHomeSet(s.config.CaldavUri, s.config.CaldavUser, s.config.CaldavPassword, homeSetUri, s.config.Calendar)
	if err != nil {
		log.Fatal("Error reading calendar URI: ", err)
	}
	log.Printf("Calendar %d:\t\t%s\n", s.config.Calendar, calendarUri)
	log.Println("----------------------------------------------------")
	log.Println("Fetching availability from CalDAV server, please wait..")

	// Get availability data for the calendar specified in the config
	calData, err := caldav.GetAvailability(s.config.CaldavUri, s.config.CaldavUser, s.config.CaldavPassword, calendarUri)
	if err != nil {
		log.Fatal("Error reading calendar URI: ", err)
	}

	err = s.UpdateAvailability(calData)
	if err != nil {
		log.Fatalf("Error fetching availability: %v", err)
	}

	log.Println("Database initialized")
	log.Printf("Next db refresh (cron): %v\n", c.Entries()[0].Schedule.Next(time.Now()).Format("2006-01-02 15:04"))
	log.Println("----------------------------------------------------")
}
